unit GenericCommand;

interface

uses
  Windows, Graphics, Classes, PCB, CharTable;

type
  TCmdParType = (ptBit,ptCoord,ptText);

type
  TCmdParameters = class (TStringList)
  private
    function GetCmdParameter(Name : string) : string;
    procedure SetCmdParameter(Name : string; Value : string);
  public
    property Parameters[Name : string] : string read GetCmdParameter write SetCmdParameter; default;
  end;

type
  TGenericCommand = class  // Basic class for all script commands
    constructor Create;
    destructor Destroy; override;
  protected
    function LoadImage : integer;
    function GetCmdText : string; virtual; abstract;
  public
    Image       : TBitmap;
    ImgName     : string;
    Description : string;
    Parameters  : TCmdParameters;
    function GetEditType : integer; virtual;
    procedure DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false); virtual;
    function Load(var Data; var Index : integer) : integer; virtual; abstract;
    procedure Save(var Data; var Index : integer); virtual; abstract;
    function Size : integer; virtual; abstract;
    function Clone : TGenericCommand; virtual; abstract;
    function Compare(ACmd : TGenericCommand) : integer; virtual; abstract;
    property AsText : string read GetCmdText;
    function Run(PCB : TPCB; CharTable : TCharTable; NoDelay : boolean) : integer; virtual; abstract;
    procedure Reset; virtual;
  end;

type
  TCommandSet = class (TList)
    constructor Create;
    procedure Clear; override;
  private
    function GetCommand(Idx : integer) : TGenericCommand;
    function GetCurrentCommand : TGenericCommand;
    function GetSize : integer;
  public
    ActiveCommand : integer;
    property CurrentCommand : TGenericCommand read GetCurrentCommand;
    property Commands[Idx : integer] : TGenericCommand read GetCommand; default;
    property Size : integer read GetSize;
    function Run(PCB : TPCB; CharTable : TCharTable; NoDelay : boolean) : integer;
    procedure Reset;
  end;

implementation

uses
  Constants, SysUtils;

// ========== TCmdParameters ===================================================

function TCmdParameters.GetCmdParameter(Name : string) : string;
begin
  if IndexOfName(uppercase(Name)) >= 0 then
    result := Values[uppercase(Name)]
  else
    result := '';
end;

// -----------------------------------------------------------------------------

procedure TCmdParameters.SetCmdParameter(Name : string; Value : string);
var
  ii : integer;
begin
  ii := IndexOfName(uppercase(Name));
  if ii < 0 then
    Add(uppercase(Name)+'='+Value)
  else
    Strings[ii] := uppercase(Name)+'='+Value;
end;

// ========== TGenericCommand ==================================================

constructor TGenericCommand.Create;
begin
  inherited Create;
  Image := TBitmap.Create;
  ImgName := '';
  Description := '';
  Parameters := TCmdParameters.Create;
end;

// -----------------------------------------------------------------------------

destructor TGenericCommand.Destroy;
begin
  Image.Free;
  Parameters.Free;
  inherited Destroy;
end;

// -----------------------------------------------------------------------------

function TGenericCommand.LoadImage : integer;
var
  MH : THandle;
begin
  MH := GetModuleHandle(nil);
  result := 0;
  if ImgName = '' then exit;
  try
    Image.LoadFromResourceName(MH,ImgName);
  except
    ImgName := '';
  end;
  result := 1;
end;

// -----------------------------------------------------------------------------

procedure TGenericCommand.DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false);
begin
  CC.Brush.Style := bsSolid;
  if Highlighted then
    CC.Brush.Color := CMD_SCREEN_COLOR
  else
    CC.Brush.Color := Dark(CMD_SCREEN_COLOR,0.5);
  CC.Pen.Color := clBlack;
  CC.Pen.Style := psDot;
  CC.Pen.Width := 0;
  CC.Rectangle(Rect);
  if ImgName = '' then exit;
  CC.StretchDraw(Rect,Image);
end;

// -----------------------------------------------------------------------------

procedure TGenericCommand.Reset;
begin
  // Empty function for those commands, which do not require qny reset
end;

// -----------------------------------------------------------------------------

function TGenericCommand.GetEditType : integer;
begin
  result := ED_NO_EDIT;
end;

// ========== TCommandSet ======================================================

constructor TCommandSet.Create;
begin
  inherited Create;
  ActiveCommand := 0;
end;

// -----------------------------------------------------------------------------

procedure TCommandSet.Clear;
var
  ii : integer;
begin
  for ii:=0 to Count-1 do Commands[ii].Free;
  ActiveCommand := 0;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

function TCommandSet.GetCommand(Idx : integer) : TGenericCommand;
begin
  if (Idx < 0) or (Idx >= Count) then
    result := nil
  else
    result := items[Idx];
end;

// -----------------------------------------------------------------------------

function TCommandSet.Run(PCB : TPCB; CharTable : TCharTable; NoDelay : boolean) : integer;
begin
  result := 0;
  if Count = 0 then exit;
  if ActiveCommand >= Count then Reset;
  result := Commands[ActiveCommand].Run(PCB,CharTable,NoDelay);
  case result of
    -1:
    begin
      Reset;
    end;
    0,2:
    begin
      inc(ActiveCommand);
      if ActiveCommand >= Count then
      begin
        result := -1;
        Reset;
      end;
    end;
  end;
end;

// -----------------------------------------------------------------------------

procedure TCommandSet.Reset;
var
  ii : integer;
begin
  ActiveCommand := 0;
  for ii:=0 to Count-1 do Commands[ii].Reset;
end;

// -----------------------------------------------------------------------------

function TCommandSet.GetCurrentCommand : TGenericCommand;
begin
  if ActiveCommand >= Count then
    result := nil
  else
    result := Commands[ActiveCommand];
end;

// -----------------------------------------------------------------------------

function TCommandSet.GetSize : integer;
var
  ii : integer;
begin
  result := 0;
  for ii:=0 to Count-1 do result := result+Commands[ii].Size;
end;

// -----------------------------------------------------------------------------

end.

