unit GenericDevice;

interface

uses
  Windows, CharTable, ScriptSet, Graphics, SerialIO, Inputs, LEDs, PCB,
  GenericCommand;

type
  TGenericDevice = class
    constructor Create(APort : string; AIDString : string);
    destructor Destroy; override;
  protected
    SIO               : TSerialIO;
    Resp              : string;
    function SIOSend(ss : string) : integer;
    procedure SIOCallback(AEvent: integer; ASender: TSerialIO);
    function WaitForEOL(MaxTout : integer = 10) : integer;
    function GetRespStr : string;
    function GetScrLimit(ScrNo : byte) : integer; virtual;
  public
    Port              : string;
    IDString          : string;
    ScriptSet         : TScriptSet;
    PCB               : TPCB;
    CharTable         : TCharTable;
    TopMem            : dword;
    Version           : string;
    SupportedCommands : TCommandSet;
    Inputs            : TInputs;
    InfoString        : string;
    ActiveScript      : integer;
    Modified          : boolean;
    LastRect          : TRect;
    property ScriptLimit[ScrNo : byte] : integer read GetScrLimit;
    function GetMaxCharSize : integer; virtual;
    function ReadDevice : integer; virtual; abstract;
    function WriteDevice : integer; virtual; abstract;
    function ReadInfo : integer; virtual; abstract;
    procedure Reset; virtual; abstract;
    procedure Clear;
    procedure DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false);
    function Run(NoDelay : boolean; CS : TCommandSet;  Cmd : TGenericCommand) : TUndo; virtual; abstract;
    procedure UpdateActiveScript; virtual; abstract;
    function WriteTo(FileName : string; ID : integer) : integer; overload; virtual; abstract;
    function WriteTo(FileName : string; Scr : TCommandSet) : integer; overload; virtual; abstract;
    function ReadFrom(FileName : string; ID : integer) : integer; overload; virtual; abstract;
    function ReadFrom(FileName : string; Scr : TCommandSet) : integer; overload; virtual; abstract;
    function GetCharTblLimits : TPoint; virtual;
    function CmdFromText(ACmdText : string) : TGenericCommand; virtual; abstract;
  end;

implementation

uses
  Constants, SysUtils;

// =============================================================================

constructor TGenericDevice.Create(APort : string; AIDString : string);
begin
  inherited Create;
  SIO := nil;
  Port := APort;
  IDString := AIDString;
  CharTable := TCharTable.Create;
  ScriptSet := TScriptSet.Create;
  PCB := TPCB.Create;
  Inputs := TInputs.Create;
  TopMem := $1000;
  ActiveScript := 0;
  SupportedCommands := TCommandSet.Create;
  Modified := false;
end;

// -----------------------------------------------------------------------------

procedure TGenericDevice.Clear;
begin
  if assigned(SIO) then SIO.Terminate;
  SIO := nil;
  if assigned(CharTable) then CharTable.Clear;
  if assigned(ScriptSet) then ScriptSet.Clear;
  if assigned(PCB) then PCB.Clear;
  Inputs.Clear;
  InfoString := '';
  TopMem := $1000;
  ActiveScript := 0;
end;

// -----------------------------------------------------------------------------

destructor TGenericDevice.Destroy;
begin
  FreeAndNil(Inputs);
  FreeAndNil(CharTable);
  FreeAndNil(ScriptSet);
  FreeAndNil(SupportedCommands);
  InfoString := '';
  if assigned(SIO) then SIO.Terminate;
  inherited Destroy;
end;

// -----------------------------------------------------------------------------

function TGenericDevice.SIOSend(ss : string) : integer;
begin
  result := 0;
  if not assigned(SIO) then exit;
  result := SIO.Send(@ss[1],length(ss));
end;

// -----------------------------------------------------------------------------

procedure TGenericDevice.SIOCallback(AEvent: integer; ASender: TSerialIO);
var
  sBuf : char;
begin
  if (AEvent = SIO_CB_EVENT_BYTE_RECEIVED) then
  begin
    if ASender.Recv(@sBuf,1) = 1 then
    begin
      if sbuf <> #0 then
      begin
        Resp := Resp+sBuf;
      end;
    end;
  end;
end;

// -----------------------------------------------------------------------------

function TGenericDevice.WaitForEOL(MaxTout : integer = 10) : integer;
var
  tout : integer;
begin
  result := 0;
  if not assigned(SIO) then exit;
  tout := MaxTout;
  while (pos(#13,Resp) = 0) and (tout > 0) do
  begin
    dec(tout);
    sleep(50);
  end;
  result := tout;
end;

// -----------------------------------------------------------------------------

function TGenericDevice.GetRespStr : string;
var
  ii,jj : integer;
begin
  if not assigned(SIO) then
  begin
    result := '';
    exit;
  end;
  setlength(result,length(Resp));
  ii := 1;
  jj := 1;
  while (ii <= length(Resp)) do
  begin
    if Resp[ii] = #13 then
    begin
      setlength(result,jj-1);
      Resp := copy(Resp,ii+1,length(Resp));
      exit;
    end
    else
    begin
      if byte(Resp[ii]) >= $20 then
      begin
        result[jj] := Resp[ii];
        inc(jj);
      end;
    end;
    inc(ii);
  end;
  result := '';
end;

// -----------------------------------------------------------------------------

procedure TGenericDevice.DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false);
var
  prect : TRect;
  irect : TRect;
begin
  LastRect := Rect;
  // The device image comprise of 2 parts:
  //   1. PCB image
  //   2. Inputs
  prect := Rect;
  irect := Rect;
  // Inputs are assigned the bottom 20% of the whole rect
  prect.Bottom := round((prect.Top+prect.Bottom)*0.8);
  irect.Top := prect.Bottom;
  PCB.DrawTo(CC,prect,Highlighted);
  Inputs.DrawTo(CC,irect,Highlighted);
end;

// -----------------------------------------------------------------------------

function TGenericDevice.GetMaxCharSize : integer;
begin
  result := 5;
end;

// -----------------------------------------------------------------------------

function TGenericDevice.GetCharTblLimits : TPoint;
begin
  result.X := $20;
  result.Y := $5F;
end;

// -----------------------------------------------------------------------------

function TGenericDevice.GetScrLimit(ScrNo : byte) : integer;
begin
  result := 0; // The limit is TopMem
end;

// -----------------------------------------------------------------------------

end.

