unit Inputs;

interface

uses
  Windows, Classes, Graphics;

type
  TGenericInput = class
    constructor Create(AName : string);
  private
    Ratio : double;
  public
    Name  : string;
    LastRect : TRect;
    LastHighlighted : boolean;
    procedure DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false); virtual; abstract;
    procedure Click(X,Y : integer); virtual; abstract;
    procedure Reset; virtual; abstract;
  end;

type
  TPinInput = class (TGenericInput)
    constructor Create(AName : string);
  public
    Value : boolean;
    procedure DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false); override;
    procedure Click(X,Y : integer); override;
    procedure Reset; override;
  end;

type
  TInputs = class (TList)
  private
    function GetInput(Idx : integer) : TGenericInput;
  public
    LastRect : TRect;
    property Inputs[Idx : integer] : TGenericInput read GetInput; default;
    procedure DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false);
    function Highlight(CC : TCanvas; X,Y : integer) : integer;
    function Click(X,Y : integer) : integer;
    procedure Reset;
  end;

implementation

uses
  Constants, Types;

// ========== TGenericInput ====================================================

constructor TGenericInput.Create(AName : string);
begin
  inherited Create;
  Name := AName;
  Ratio := 1;
  LastHighlighted := false;
end;

// ========== TPinInput ========================================================

constructor TPinInput.Create(AName : string);
begin
  inherited Create(AName);
  Value := false;
end;

// -----------------------------------------------------------------------------

procedure TPinInput.DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false);
var
  r : TRect;
begin
  LastRect := Rect;
  r.Top := Rect.Top+2;
  r.Bottom := Rect.Bottom-2;
  r.Left := Rect.Left+2;
  r.Right := r.Left+(r.Bottom-r.Top);
  if Highlighted then
    CC.Pen.Color := INP_BTN_BORDER
  else
    CC.Pen.Color := Dark(INP_BTN_BORDER,0.8);
  CC.Pen.Width := 5;
  CC.Ellipse(r);
  if Value then
    CC.Pen.Color := clYellow
  else
    CC.Pen.Color := Dark(clYellow,0.6);
  r.Left := (r.Left+r.Right-7) div 2;
  r.Top := (r.Top+r.Bottom-7) div 2;
  r.Right := r.Left+8;
  r.Bottom := r.Top+8;
  CC.Ellipse(r);
end;

// -----------------------------------------------------------------------------

procedure TPinInput.Click(X: Integer; Y: Integer);
begin
  Value := not Value;
end;

// -----------------------------------------------------------------------------

procedure TPinInput.Reset;
begin
  Value := false;
end;

// ========== TInputs ==========================================================

procedure TInputs.DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false);
var
  Ratio : double;
  ii : integer;
  MaxWidth : integer;
  MaxHeight : integer;
  prect : TRect;
  irect : TRect;
begin
  LastRect := Rect;

  // We must fill the background with a color first.
  CC.Brush.Style := bsSolid;
  CC.Brush.Color := INP_SCREEN_COLOR;
  CC.Pen.Style := psSolid;
  CC.Pen.Color := Dark(INP_SCREEN_COLOR,0.6);
  CC.Pen.Width := 2;
  CC.Rectangle(Rect);

  //Calculate total ratio
  Ratio := 0;
  for ii:=0 to Count-1 do
    Ratio := Ratio+Inputs[ii].Ratio;

  MaxHeight := 28;
  MaxWidth := round(MaxHeight*Ratio);

  prect.Left := (Rect.Left+Rect.Right-MaxWidth) div 2;
  prect.Right := prect.Left+MaxWidth;
  prect.Top := (Rect.Top+Rect.Bottom-MaxHeight) div 2;
  if prect.Top < (Rect.Top+5) then prect.Top := Rect.Top+5;
  prect.Bottom := prect.Top+MaxHeight;

  // Draw the input plate
  CC.Brush.Color := Dark(INP_COLOR,0.4);
  CC.Pen.Style := psSolid;
  CC.Pen.Color := INP_BORDER;
  CC.Pen.Width := 2;
  CC.Rectangle(prect);

  for ii:=0 to Count-1 do
  begin
    if ii = 0 then
      irect.Left := prect.Left+2
    else
      irect.Left := irect.Right+4;
    irect.Top := prect.Top+2;
    irect.Bottom := prect.Bottom-2;
    irect.Right := irect.Left+round((irect.Bottom-irect.Top)*Inputs[ii].Ratio);
    Inputs[ii].DrawTo(CC,irect,Inputs[ii].LastHighLighted);
  end;

  CC.Pen.Style := psSolid;
  CC.Pen.Color := clWhite;
  CC.Pen.Width := 1;
  CC.MoveTo(prect.Left-1,prect.Bottom);
  CC.LineTo(prect.Left-1,prect.Top-1);
  CC.LineTo(prect.Right,prect.Top-1);
  CC.Pen.Color := clGray;
  CC.LineTo(prect.Right,prect.Bottom);
  CC.LineTo(prect.Left-1,prect.Bottom);
end;

// -----------------------------------------------------------------------------

function TInputs.GetInput(Idx : integer) : TGenericInput;
begin
  if (Idx < 0) or (Idx >= Count) then
    result := nil
  else
    result := items[Idx];
end;

// -----------------------------------------------------------------------------

function TInputs.Highlight(CC : TCanvas; X,Y : integer) : integer;
var
  ii : integer;
begin
  result := -1;
  for ii:=0 to Count-1 do
  begin
    if (X > Inputs[ii].LastRect.Left) and (X < Inputs[ii].LastRect.Right) and
       (Y > Inputs[ii].LastRect.Top) and (Y < Inputs[ii].LastRect.Bottom) then
    begin
      if not Inputs[ii].LastHighlighted then
        Inputs[ii].DrawTo(CC,Inputs[ii].LastRect,true);
      Inputs[ii].LastHighlighted := true;
      result := ii;
    end
    else
    begin
      if Inputs[ii].LastHighlighted then
        Inputs[ii].DrawTo(CC,Inputs[ii].LastRect,false);
      Inputs[ii].LastHighlighted := false;
    end;
  end;
end;

// -----------------------------------------------------------------------------

function TInputs.Click(X,Y : integer) : integer;
var
  ii : integer;
begin
  result := 0;
  for ii:=0 to Count-1 do
  begin
    if (X >= Inputs[ii].LastRect.Left) and (X <= Inputs[ii].LastRect.Right) and
       (Y >= Inputs[ii].LastRect.Top) and (Y <= Inputs[ii].LastRect.Bottom) then
    begin
      Inputs[ii].Click(X,Y);
      result := 1;
      break;
    end;
  end;
end;

// -----------------------------------------------------------------------------

procedure TInputs.Reset;
var
  ii : integer;
begin
  for ii:=0 to Count-1 do Inputs[ii].Reset;
end;

// -----------------------------------------------------------------------------

end.

