unit PCB;

interface

uses
  Windows, Graphics, Classes, LEDs;

type
  TUndo = array [0..999,0..999] of boolean;
  PUndo = ^TUndo;

type
  TLEDRow = class (TList)
    constructor Create;
    procedure Clear; override;
  private
    function GetLED(Col : integer) : TLED;
    function GetColCount : integer;
    procedure SetColCount(Value : integer);
  public
    property ColCount : integer read GetColCount write SetColCount;
    property LEDs[Col : integer] : TLED read GetLED; default;
    procedure Assign(ALEDRow : TLEDRow); reintroduce;
  end;

type
  TPCB = class (TList)  // Stored by rows
    constructor Create;
    procedure Clear; override;
  private
    function GetColCount : integer;
    function GetRowCount : integer;
    procedure SetRowCount(Value : integer);
    procedure SetColCount(Value : integer);
    function GetLED(Col : integer; Row : integer) : TLED;
    function GetLEDRow(Row : integer) : TLEDRow;
  public
    LastRect : TRect;
    // The PCB object just stores the follofing variables
    Tag1 : integer;              // Character delay
    Tag2 : integer;
    Tag3 : integer;              // Character Spacing
    Tag4 : string;               // Horizontal buffer
    Tag5 : byte;                 // Vertical buffer
    Tag6 : integer;              // Direction
    Tag7 : integer;              // Mirroring
    property RowCount : integer read GetRowCount write SetRowCount;
    property ColCount : integer read GetColCount write SetColCount;
    property LEDArray[Col : integer; Row : integer] : TLED read GetLed; default;
    property LEDRows[Row : integer] : TLEDRow read GetLEDRow;
    procedure DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false);
    procedure Off;
  end;

implementation

uses
  Constants, Types;

// ========== TLEDRow ==========================================================

constructor TLEDRow.Create;
begin
  inherited Create;
  Add(TLED.Create);
end;

// -----------------------------------------------------------------------------

procedure TLEDRow.Clear;
var
  ii : integer;
begin
  for ii := 0 to Count - 1 do
    if assigned(LEDs[ii]) then LEDs[ii].Free;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

function TLEDRow.GetLED(Col : integer) : TLED;
begin
  if (Col < 0) or (Col >= Count) then
    result := nil
  else
    result := items[Col];
end;

// -----------------------------------------------------------------------------

function TLEDRow.GetColCount : integer;
begin
  result := Count;
end;

// -----------------------------------------------------------------------------

procedure TLEDRow.SetColCount(Value : integer);
begin
  while Count > Value do
  begin
    LEDs[Count-1].Free;
    Delete(Count-1);
  end;
  while Count < Value do Add(TLED.Create);
  if Count = 0 then Add(TLED.Create);
end;

// -----------------------------------------------------------------------------

procedure TLEDRow.Assign(ALEDRow : TLEDRow);
var
  ii : integer;
begin
  ColCount := ALEDRow.ColCount;
  for ii:=0 to ColCount-1 do LEDs[ii].Assign(ALEDRow[ii]);
end;

// ========== TPCB =============================================================

procedure TPCB.DrawTo(CC : TCanvas; Rect : TRect; Highlighted : boolean = false);
var
  MaxWidth  : integer;
  MaxHeight : integer;
  f1,f2     : double;
  prect     : TRect;
  lrect     : TRect;
  iCol,iRow : integer;
  Row       : TLEDRow;
  fLEDSize  : double;
begin
  LastRect := Rect;

  // We must fill the background with a color first.
  CC.Brush.Style := bsSolid;
  CC.Brush.Color := PCB_SCREEN_COLOR;
  CC.Pen.Style := psSolid;
  CC.Pen.Color := Dark(PCB_SCREEN_COLOR,0.6);
  CC.Pen.Width := 2;
  CC.Rectangle(Rect);

  // Calculate the MaxWidth and MaxHeight of the display rectangle
  f1 := ColCount/RowCount;
  f2 := (Rect.Right-Rect.Left)/(Rect.Bottom-Rect.Top);
  if f1 >= f2 then
  begin
    MaxWidth := Rect.Right-Rect.Left-20; // 10 pixels of space on each side
    fLEDSize := MaxWidth / ColCount;
    if fLEDSize > 80 then
    begin
      fLEDSize := 80;
      MaxWidth := trunc(ColCount*fLEDSize)+1;
    end;
    MaxHeight := round(MaxWidth/f1);
  end
  else
  begin
    MaxHeight := Rect.Bottom-Rect.Top-20; // 10 pixels of space on each side
    fLEDSize := MaxHeight / RowCount;
    if fLEDSize > 80 then
    begin
      fLEDSize := 80;
      MaxHeight := trunc(RowCount*fLEDSize)+1;
    end;
    MaxWidth := round(MaxHeight*f1);
  end;
  prect.Left := (Rect.Left+Rect.Right-MaxWidth) div 2;
  prect.Right := prect.Left+MaxWidth;
  prect.Top := (Rect.Top+Rect.Bottom-MaxHeight) div 2;
  prect.Bottom := prect.Top+MaxHeight;

  // Draw the board
  CC.Brush.Color := Dark(PCB_COLOR,0.2);
  CC.Pen.Style := psSolid;
  CC.Pen.Color := PCB_BORDER;
  CC.Pen.Width := 2;
  CC.Rectangle(prect);

  for iRow := 0 to RowCount - 1 do
  begin
    Row := LEDRows[iRow];
    for iCol := 0 to Row.ColCount-1 do
    begin
      lrect.Left := round(prect.Left+iCol*fLEDSize);
      lrect.Right := round(lrect.Left+fLEDSize);
      lrect.Top := round(prect.Top+iRow*fLEDSize);
      lrect.Bottom := round(lrect.Top+fLEDSize);
      Row[iCol].DrawTo(CC,lrect,HighLighted);
    end;
  end;

  CC.Pen.Style := psSolid;
  CC.Pen.Color := clWhite;
  CC.Pen.Width := 1;
  CC.MoveTo(prect.Left-1,prect.Bottom);
  CC.LineTo(prect.Left-1,prect.Top-1);
  CC.LineTo(prect.Right,prect.Top-1);
  CC.Pen.Color := clGray;
  CC.LineTo(prect.Right,prect.Bottom);
  CC.LineTo(prect.Left-1,prect.Bottom);
end;

// -----------------------------------------------------------------------------

constructor TPCB.Create;
begin
  inherited Create;
  Add(TLEDRow.Create);
end;

// -----------------------------------------------------------------------------

procedure TPCB.Clear;
var
  ii : integer;
begin
  for ii := 0 to Count - 1 do
    if assigned(LEDRows[ii]) then LEDRows[ii].Free;
  inherited Clear;
  Add(TLEDRow.Create);
end;

// -----------------------------------------------------------------------------

function TPCB.GetColCount : integer;
begin
  result := LEDRows[0].ColCount;
end;

// -----------------------------------------------------------------------------

function TPCB.GetRowCount : integer;
begin
  result := Count;
end;

// -----------------------------------------------------------------------------

procedure TPCB.SetRowCount(Value : integer);
begin
  while Count > Value do
  begin
    LEDRows[Count-1].Free;
    Delete(Count-1);
  end;
  while Count < Value do Add(TLEDRow.Create);
  if Count = 0 then Add(TLEDRow.Create);
end;

// -----------------------------------------------------------------------------

procedure TPCB.SetColCount(Value : integer);
var
  ii : integer;
begin
  for ii := 0 to Count-1 do
    LEDRows[ii].ColCount := Value;
end;

// -----------------------------------------------------------------------------

function TPCB.GetLED(Col : integer; Row : integer) : TLED;
begin
  if (Row < 0) or (Row >= RowCount) then
    result := nil
  else
  if (Col < 0) or (Col >= ColCount) then
    result := nil
  else
    result := LEDRows[Row][Col];
end;

// -----------------------------------------------------------------------------

function TPCB.GetLEDRow(Row : integer) : TLEDRow;
begin
  if (Row < 0) or (Row >= RowCount) then
    result := nil
  else
    result := items[Row];
end;

// -----------------------------------------------------------------------------

procedure TPCB.Off;
var
  row,col : integer;
begin
  for row:=0 to RowCount-1 do
    for col:=0 to ColCount-1 do
      LEDArray[col,row].Active := false;
end;

// -----------------------------------------------------------------------------

end.

