//;;; --------------------------------------------
//;
//; Communication Command List
//;
//; 10						Get Device ID And Firmware
//; 2x						Download Script Number x
//; 3x yy CRC  '64dataBlock CRC' * yy		Upload Script Number x
//; 40						Download Char Table
//; 50 yy	CRC  '64dataBlock CRC' * yy		Upload Char Table

unit SLV01;

interface

uses
  GenericDevice, GenericCommand, PCB;

type
  TSLV01_Block = string[130];

type
  TSLV01_BlockSet = array of TSLV01_Block;

type
  TSLV01 = class(TGenericDevice)
    constructor Create(APort : string; AIDString : string);
    function GetScrLimit(ScrNo : byte) : integer; override;
  private
    function PrepareScript(CS1,CS2 : TCommandSet) : TSLV01_BlockSet;
    function PrepareCharTbl : TSLV01_BlockSet;
  public
    function ReadDevice : integer; override;
    function WriteDevice : integer; override;
    function ReadInfo : integer; override;
    procedure Reset; override;
    procedure UpdateActiveScript; override;
    function Run(NoDelay : boolean; CS : TCommandSet;  Cmd : TGenericCommand) : TUndo; override;
    function WriteTo(FileName : string; ID : integer) : integer; override;
    function WriteTo(FileName : string; Scr : TCommandSet) : integer; overload; override;
    function ReadFrom(FileName : string; ID : integer) : integer; override;
    function ReadFrom(FileName : string; Scr : TCommandSet) : integer; overload; override;
    function CmdFromText(ACmdText : string) : TGenericCommand; override;
  end;

implementation

uses
  SysUtils, SerialIO, SLV01Commands, CharTable, Inputs, Constants, Dialogs,
  Classes, Forms, Windows, Main;

// Default scripts
var
  Script0000s : array [1..4] of byte = ($cf,$02,$0c,$00);
  Script0001s : array [1..32] of byte = ($0f,$02,$c1,$08,$c3,$0f,$c1,$08,$c3,$0f,
                $c1,$08,$c3,$0f,$c1,$08,$03,$c2,$0f,$88,byte(' '),byte('S'),
                byte('T'),byte('O'),byte('P'),byte(' '),byte(' '),byte(' '),$02,
                $0f,$00,$00);
  Script0001e : array [1..2] of byte = ($01,$01);
  Script0010s : array [1..15] of byte = ($03,$0f,$44,$4e,$5f,$4e,$c1,$4e,$4e,$4e,
                $4e,$c1,$4e,$4e,$00);
  Script0010e : array [1..13] of byte = ($40,$40,$40,$40,$c1,$40,$40,$40,$40,$c1,
                $40,$40,$01);
  Script0011s : array [1..32] of byte = ($0f,$02,$c1,$08,$c3,$0f,$c1,$08,$c3,$0f,
                $c1,$08,$c3,$0f,$c1,$08,$03,$c2,$0f,$88,byte(' '),byte('S'),
                byte('T'),byte('O'),byte('P'),byte(' '),byte(' '),byte(' '),$02,
                $0f,$00,$00);
  Script0011e : array [1..2] of byte = ($01,$01);
  Script0100s : array [1..15] of byte = ($03,$0f,$64,$6e,$7f,$6e,$c1,$6e,$6e,$6e,
                $6e,$c1,$6e,$6e,$00);
  Script0100e : array [1..13] of byte = ($60,$60,$60,$60,$c1,$60,$60,$60,$60,$c1,
                $60,$60,$01);
  Script0101s : array [1..32] of byte = ($0f,$02,$c1,$08,$c3,$0f,$c1,$08,$c3,$0f,
                $c1,$08,$c3,$0f,$c1,$08,$03,$c2,$0f,$88,byte(' '),byte('S'),
                byte('T'),byte('O'),byte('P'),byte(' '),byte(' '),byte(' '),$02,
                $0f,$00,$00);
  Script0101e : array [1..2] of byte = ($01,$01);
  Script0110s : array [1..2] of byte = ($01,$01);
  Script0110e : array [1..2] of byte = ($01,$01);
  Script0111s : array [1..32] of byte = ($0f,$02,$c1,$08,$c3,$0f,$c1,$08,$c3,$0f,
                $c1,$08,$c3,$0f,$c1,$08,$03,$c2,$0f,$88,byte(' '),byte('S'),
                byte('T'),byte('O'),byte('P'),byte(' '),byte(' '),byte(' '),$02,
                $0f,$00,$00);
  Script0111e : array [1..2] of byte = ($01,$01);

// =============================================================================

function AddToCRC(CRCIn,ByteIn : byte) : byte;
begin
  if (CRCIn and $80) = 0 then
    result := byte(byte(CRCIn shl 1) + byte(ByteIn))
  else
    result := byte(byte((CRCIn shl 1) or $01) + byte(ByteIn));
end;

// ========== TSLV01 ===========================================================

constructor TSLV01.Create(APort : string; AIDString : string);
begin
  inherited Create(APort,AIDString);
  TopMem := $2000;
  PCB.RowCount := 5;
  PCB.ColCount := 10;
  SupportedCommands.Add(TSLV01Cmd_Stop.Create);
  SupportedCommands.Add(TSLV01Cmd_Restart.Create);
  SupportedCommands.Add(TSLV01Cmd_SetAll.Create);
  SupportedCommands.Add(TSLV01Cmd_ClearAll.Create);
  SupportedCommands.Add(TSLV01Cmd_InsFromTop.Create);
  SupportedCommands.Add(TSLV01Cmd_InsFromBottom.Create);
  SupportedCommands.Add(TSLV01Cmd_SetPixel.Create);
  SupportedCommands.Add(TSLV01Cmd_ClearPixel.Create);
  SupportedCommands.Add(TSLV01Cmd_SetBrightness.Create);
  SupportedCommands.Add(TSLV01Cmd_ScrollRowLeft.Create);
  SupportedCommands.Add(TSLV01Cmd_ScrollRowRight.Create);
  SupportedCommands.Add(TSLV01Cmd_InsFromLeft.Create);
  SupportedCommands.Add(TSLV01Cmd_InsFromRight.Create);
  SupportedCommands.Add(TSLV01Cmd_ScrollMsg.Create);
  SupportedCommands.Add(TSLV01Cmd_Delay.Create);
  Inputs.Add(TPinInput.Create('Input #1 (Left Turn)'));
  Inputs.Add(TPinInput.Create('Input #2 (Stop)'));
  Inputs.Add(TPinInput.Create('Input #3 (Right Turn)'));
end;

// -----------------------------------------------------------------------------

function TSLV01.PrepareCharTbl : TSLV01_BlockSet;
var
  Buf : PByteArray;
  ch  : TCharacter;
  bb  : byte;
  ii,jj : integer;
  ss : string[130];
  sb : string;
  CRC : byte;
begin
  setlength(result,6);
  GetMem(Buf,384);
  try
    for ii:=0 to 383 do Buf[ii] := 0;
    for ii:=0 to 63 do
    begin
      ch := CharTable[32+ii];
      bb := length(ch.Data);
      Buf[ii*6] := bb;
      if bb > 0 then Buf[ii*6+1] := ch.Data[0];
      if bb > 1 then Buf[ii*6+2] := ch.Data[1];
      if bb > 2 then Buf[ii*6+3] := ch.Data[2];
      if bb > 3 then Buf[ii*6+4] := ch.Data[3];
      if bb > 4 then Buf[ii*6+5] := ch.Data[4];
    end;
    for ii:=0 to 5 do
    begin
      CRC := 0;
      ss[0] := #128;
      for jj:=0 to 63 do
      begin
        CRC := AddToCRC(CRC,Buf[(ii*64)+jj]);
        sb := HexByte(Buf[(ii*64)+jj]);
        ss[jj*2+1] := sb[1];
        ss[jj*2+2] := sb[2];
      end;
      sb := HexByte(CRC);
      ss := ss+sb;
      result[ii] := ss;
    end;
  finally
    FreeMem(Buf);
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.PrepareScript(CS1,CS2 : TCommandSet) : TSLV01_BlockSet;
var
  Buf : PByteArray;
  Index : integer;
  ii,jj : integer;
  cmd : TGenericCommand;
  BlkCnt : integer;
  ss : string[130];
  sb : string;
  CRC : byte;
begin
  setlength(result,0);
  GetMem(Buf,2048);
  try
    for ii:=0 to 2047 do Buf[ii] := $ff;
    Index := 0;
    for ii:=0 to CS1.Count-1 do
    begin
      cmd := CS1[ii];
      if (Cmd.Size+Index) > 1024 then exit;
      cmd.Save(Buf^,Index);
      if (cmd is TSLV01Cmd_Stop) or (cmd is TSLV01Cmd_Restart) then break;
    end;
    BlkCnt := Index div 64;
    if BlkCnt > 8 then exit;
    setlength(result,8);
    if (Index mod 64) > 0 then inc(BlkCnt);
    for ii:=0 to BlkCnt-1 do
    begin
      CRC := 0;
      ss[0] := #128;
      for jj:=0 to 63 do
      begin
        CRC := AddToCRC(CRC,Buf[(ii*64)+jj]);
        sb := HexByte(Buf[(ii*64)+jj]);
        ss[jj*2+1] := sb[1];
        ss[jj*2+2] := sb[2];
      end;
      sb := HexByte(CRC);
      ss := ss+sb;
      result[ii] := ss;
    end;
    for ii:=BlkCnt to 7 do
    begin
      CRC := 0;
      ss[0] := #128;
      for jj:=0 to 63 do
      begin
        CRC := AddToCRC(CRC,$ff);
        ss[jj*2+1] := 'F';
        ss[jj*2+2] := 'F';
      end;
      sb := HexByte(CRC);
      ss := ss+sb;
      result[ii] := ss;
    end;
    if not assigned(CS2) then exit;
    // ---------- Processing CS2 ----------
    for ii:=0 to 2047 do Buf[ii] := $ff;
    Index := 0;
    for ii:=0 to CS2.Count-1 do
    begin
      cmd := CS2[ii];
      if (Cmd.Size+Index) > 1024 then exit;
      cmd.Save(Buf^,Index);
      if (cmd is TSLV01Cmd_Stop) or (cmd is TSLV01Cmd_Restart) then break;
    end;
    BlkCnt := Index div 64;
    if BlkCnt > 8 then
    begin
      setlength(result,0);
      exit;
    end;
    if (Index mod 64) > 0 then inc(BlkCnt);
    setlength(result,8+BlkCnt);
    for ii:=0 to BlkCnt-1 do
    begin
      CRC := 0;
      ss[0] := #128;
      for jj:=0 to 63 do
      begin
        CRC := AddToCRC(CRC,Buf[(ii*64)+jj]);
        sb := HexByte(Buf[(ii*64)+jj]);
        ss[jj*2+1] := sb[1];
        ss[jj*2+2] := sb[2];
      end;
      sb := HexByte(CRC);
      ss := ss+sb;
      result[ii+8] := ss;
    end;
  finally
    FreeMem(Buf);
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.ReadDevice : integer;
var
  ii  : byte;
  CRC : byte;
  bb  : byte;
  Buf : PByteArray;
  Att : integer;
  bsz : word;
  tout: integer;
  cs  : TCommandSet;
  Code : byte;
  cmd  : TGenericCommand;
  Index: integer;
  cc  : string;
  sz,b1,b2,b3,b4,b5 : byte;
  // -----------------------
  procedure AddToCC(b : byte);
  var
    ss : string;
  begin
    if b in [$55,13,10] then
      CC := CC+char(b)
    else
    begin
      ss := HexByte(b);
      CC := CC+ss;
    end;
    CRC := AddToCRC(CRC,b);
  end;
  // ------------------------------------------------
  procedure AddAChar(Code,Len,b0,b1,b2,b3,b4 : byte);
  var
    ch : TCharacter;
    ii : integer;
  begin
    ch := CharTable[Code];
    if Len = 0 then exit;
    setlength(ch.Data,Len);
    for ii:=0 to Len-1 do
    begin
      case ii of
        0: ch.Data[ii] := b0;
        1: ch.Data[ii] := b1;
        2: ch.Data[ii] := b2;
        3: ch.Data[ii] := b3;
        4: ch.Data[ii] := b4;
      end;
    end;
  end;
  // ------------------------------------------------
begin
  result := 0;
  SIO := TSerialIO.Create(Port,SIOCallback);
  try
    tout := 0;
    while (tout < 4) and (SIO.Error = SIO_ERR_CLOSED) do
    begin
      sleep(50);
      inc(tout);
    end;
    if SIO.Error <> SIO_ERR_NO_ERROR then exit;

    Att := 0;
    ii := 0; // Reading 7 scripts
    while ii < 8 do
    begin
      if Att > 0 then
        ShowProgress(0,8,ii,'Reading script '+IntToStr(ii+1)+' out of 8 (Att #'+IntToStr(Att)+')')
      else
        ShowProgress(0,8,ii,'Reading script '+IntToStr(ii+1)+' out of 8');
      Resp := '';
      CRC := 0;
      cc := '';
      AddToCC(byte('U'));
      AddToCC($e0);
      AddToCC($20 or ii);
      AddToCC(CRC);
      AddToCC(13);
      SIO.Send(@cc[1],length(cc));
      if WaitForEOL(80) = 0 then
      begin
        inc(Att);
        if Att > 2 then
        begin
          inc(ii);
          Att := 0;
          Continue;
        end;
        // Another attempt
        SIO.Send('?',1); // Try to reset...
        if WaitForEOL = 0 then
        begin
          inc(ii);
          Att := 0;
        end;
        Continue;
      end;
      CRC := 0;
      if copy(Resp,1,1) = 'E' then
      begin
        inc(Att); // Error response
        Continue;
      end;
      // Get the size part (should be $0400)
      if length(Resp) < 3 then
      begin
        inc(Att); // Result string is too short
        Continue;
      end;
      bsz := UnhexWord(copy(Resp,1,4));
      if bsz >= TopMem then
      begin
        inc(Att); // Block size is too big
        Continue;
      end;
      if ((bsz*2)+6) > length(Resp) then
      begin
        inc(Att); // Wrong number of bytes received
        Continue;
      end;
      GetMem(Buf,bsz*2);
      try
        for Index:=0 to bsz-1 do Buf[Index] := 0;
        CRC := AddToCRC(CRC,hi(bsz));
        CRC := AddToCRC(CRC,lo(bsz));
        for Index:=0 to bsz-1 do
        begin
          Buf[Index] := UnhexByte(copy(Resp,(5+(Index*2)),2));
          CRC := AddToCRC(CRC,Buf[Index]);
        end;
        bb := UnhexByte(copy(Resp,((bsz*2)+5),2));
        if bb <> CRC then
        begin
          inc(Att); // CRC error
          Continue;
        end;
        // All OK - load the script into CommandSet
        Code := ii shl 1;
        cs := ScriptSet[Code];
        cs.Clear;
        Index := 0;
        while Index < bsz do
        begin
          cmd := TSLV01_LoadCommand(Buf^,Index);
          cs.Add(cmd);
          if (cmd is TSLV01Cmd_Stop) or (cmd is TSLV01Cmd_Restart) then
          begin
            if ((Code and 1) = 0) and(ii > 0) then
            begin
              inc(Code);
              cs := ScriptSet[Code];
              cs.Clear;
              Index := $200;
            end
            else
            begin
              break;
            end;
          end;
        end;
        inc(ii);
        Att := 0;
      finally
        FreeMem(Buf);
      end;
    end;

    // Reading Character Table
    ShowProgress(0,8,8,'Reading the Character Table...');
    Resp := '';
    SIO.Send('UE04055'#13,8);
    if WaitForEOL(80) = 0 then
    begin
      ShowMessage('Timeout!');
      exit;
    end;
    CRC := 0;
    if copy(Resp,1,1) = 'E' then
    begin
      ShowMessage('Error response received: '+Resp);
    end;
    // Get the size part
    if length(Resp) < 3 then
    begin
      ShowMessage('Response of incorrect length returned');
      exit;
    end;
    bsz := UnhexWord(copy(Resp,1,4));
    if bsz <> $180 then
    begin
      ShowMessage('Response of incorrect length returned (1)');
      exit;
    end;
    if ((bsz*2)+6) > length(Resp) then
    begin
      ShowMessage('Response of incorrect length returned (2)');
      exit;
    end;
    GetMem(Buf,bsz*2);
    try
      for Index:=0 to bsz-1 do Buf[Index] := 0;
      CRC := AddToCRC(CRC,hi(bsz));
      CRC := AddToCRC(CRC,lo(bsz));
      for Index:=0 to bsz-1 do
      begin
        Buf[Index] := UnhexByte(copy(Resp,(5+(Index*2)),2));
        CRC := AddToCRC(CRC,Buf[Index]);
      end;
      bb := UnhexByte(copy(Resp,((bsz*2)+5),2));
      if bb <> CRC then
      begin
        ShowMessage('CRC Error');
        exit;
      end;
      // All OK - load characters
      for Index:=0 to 63 do
      begin
        sz := Buf[Index*6];
        b1 := Buf[Index*6+1];
        b2 := Buf[Index*6+2];
        b3 := Buf[Index*6+3];
        b4 := Buf[Index*6+4];
        b5 := Buf[Index*6+5];
        AddAChar(32+Index,sz,b1,b2,b3,b4,b5);
      end;
    finally
      FreeMem(Buf);
    end;
    result := 1;
  finally
    HideProgress;
    SIO.Terminate;
    SIO := nil;
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.WriteDevice : integer;
var
  BS  : TSLV01_BlockSet;
  ii  : byte;
  jj  : byte;
  CRC : byte;
  cc  : string;
  ss  : string;
  tout: integer;
  // -----------------------
  procedure AddToCC(b : byte);
  var
    ss : string;
  begin
    if b in [$55,13,10] then
      CC := CC+char(b)
    else
    begin
      ss := HexByte(b);
      CC := CC+ss;
    end;
    CRC := AddToCRC(CRC,b);
  end;
  // -----------------------
begin
  result := 0;
  SIO := TSerialIO.Create(Port,SIOCallback);
  try
    tout := 0;
    while (tout < 4) and (SIO.Error = SIO_ERR_CLOSED) do
    begin
      sleep(50);
      inc(tout);
    end;
    if SIO.Error <> SIO_ERR_NO_ERROR then exit;
    for ii:=0 to 7 do
    begin
      ShowProgress(0,8,ii,'Writing script '+IntToStr(ii+1)+' out of 8');
      if ii = 0 then
        BS := PrepareScript(ScriptSet[ii shl 1],nil)
      else
        BS := PrepareScript(ScriptSet[ii shl 1],ScriptSet[(ii shl 1)+1]);
      CC := '';
      CRC := 0;
      AddToCC(byte('U'));
      AddToCC($e0);
      AddToCC($30+ii);
      jj := length(BS);
      AddToCC(jj);
      AddToCC(CRC);
      Resp := '';
      SIO.Send(@CC[1],length(CC));
      if WaitForEOL > 0 then
      begin
        ShowMessage('Error returned: '+Resp);
        exit; // Some kind of error returned
      end;
      for jj:=0 to length(BS)-1 do
      begin
        Resp := '';
        SIO.Send(@BS[jj][1],length(BS[jj]));
        if WaitForEOL(160) = 0 then
        begin
          ShowMessage('Timeout at script '+IntToStr(ii)+', block '+IntToStr(jj));
          exit;
        end;
        ss := format('E03%u%2.2X',[ii,jj]);
        if copy(Resp,1,length(ss)) <> ss then
        begin
          ShowMessage('Unexpected response: '+Resp);
          exit;
        end;
      end;
      Resp := '';
      SIO.Send(#13,1);
      if (WaitForEOL(10) = 0) or (length(Resp) < 2) or (pos(' OK',Resp) = 0) then
      begin
         if Resp = '' then
          ShowMessage('No final confirmation received')
         else
          ShowMessage('Incorrect final confirmation received: '+Resp);
        exit;
      end;
    end;

    // Writing the character table
    ShowProgress(0,8,8,'Writing the Character Table...');
    BS := PrepareCharTbl;
    CC := '';
    CRC := 0;
    AddToCC(byte('U'));
    AddToCC($e0);
    AddToCC($50);
    jj := length(BS);
    AddToCC(jj);
    AddToCC(CRC);
    Resp := '';
    SIO.Send(@CC[1],length(CC));
    if WaitForEOL > 0 then
    begin
      ShowMessage('Error returned: '+Resp);
      exit; // Some kind of error returned
    end;
    for jj:=0 to length(BS)-1 do
    begin
      Resp := '';
      SIO.Send(@BS[jj][1],length(BS[jj]));
      if WaitForEOL(160) = 0 then
      begin
        ShowMessage('Timeout');
        exit;
      end;
      ss := format('E050%2.2X',[jj]);
      if copy(Resp,1,length(ss)) <> ss then
      begin
        ShowMessage('Unexpected response: '+Resp);
        exit;
      end;
    end;
    Resp := '';
    SIO.Send(#13,1);
    if (WaitForEOL(10) = 0) or (length(Resp) < 2) or (pos(' OK',Resp) = 0) then
    begin
       if Resp = '' then
        ShowMessage('No final confirmation received')
       else
        ShowMessage('Incorrect final confirmation received: '+Resp);
      exit;
    end;
    result := 1;
  finally
    HideProgress;
    SIO.Terminate;
    SIO := nil;
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.ReadInfo : integer;
begin
  // This device does not support the ReadInfo functionality
  result := 1;
end;

// -----------------------------------------------------------------------------

procedure TSLV01.Reset;
  // ------------------------------------------------
  procedure AddAChar(Code,Len,b0,b1,b2,b3,b4 : byte);
  var
    ch : TCharacter;
    ii : integer;
  begin
    ch := CharTable[Code];
    if Len = 0 then exit;
    setlength(ch.Data,Len);
    for ii:=0 to Len-1 do
    begin
      case ii of
        0: ch.Data[ii] := b0;
        1: ch.Data[ii] := b1;
        2: ch.Data[ii] := b2;
        3: ch.Data[ii] := b3;
        4: ch.Data[ii] := b4;
      end;
    end;
  end;
  // ------------------------------------------------
  procedure AddScript(Code : byte; var Data; Len : integer);
  var
    cs : TCommandSet;
    ii : integer;
  begin
    while ScriptSet.Count <= Code do ScriptSet.Add(TCommandSet.Create);
    cs := ScriptSet[Code];
    ii := 0;
    while ii < Len do
      cs.Add(TSLV01_LoadCommand(Data,ii));
  end;
  // ------------------------------------------------
begin
  // Creating a default font
  CharTable.Clear;
  AddAChar(32,$03,$00,$00,$00,$00,$00);
  AddAChar(33,$03,$00,$1d,$00,$00,$00);
  AddAChar(34,$04,$00,$18,$00,$18,$00);
  AddAChar(35,$05,$0b,$0e,$1b,$0e,$1a);
  AddAChar(36,$05,$08,$15,$1f,$15,$02);
  AddAChar(37,$05,$19,$1a,$04,$0b,$13);
  AddAChar(38,$04,$0a,$15,$17,$0b,$00);
  AddAChar(39,$03,$00,$18,$00,$00,$00);
  AddAChar(40,$03,$0e,$11,$00,$00,$00);
  AddAChar(41,$03,$11,$0e,$00,$00,$00);
  AddAChar(42,$05,$05,$06,$1c,$06,$05);
  AddAChar(43,$05,$04,$04,$1f,$04,$04);
  AddAChar(44,$03,$01,$02,$00,$00,$00);
  AddAChar(45,$04,$04,$04,$04,$04,$00);
  AddAChar(46,$03,$00,$02,$00,$00,$00);
  AddAChar(47,$05,$01,$02,$04,$08,$10);
  AddAChar(48,$04,$0e,$11,$11,$0e,$00);
  AddAChar(49,$03,$04,$08,$1f,$00,$00);
  AddAChar(50,$04,$11,$13,$15,$09,$00);
  AddAChar(51,$04,$11,$15,$15,$0a,$00);
  AddAChar(52,$04,$02,$06,$0a,$1f,$00);
  AddAChar(53,$04,$1d,$15,$15,$17,$00);
  AddAChar(54,$04,$0e,$15,$15,$02,$00);
  AddAChar(55,$04,$11,$12,$14,$18,$00);
  AddAChar(56,$04,$0a,$15,$15,$0a,$00);
  AddAChar(57,$04,$1d,$15,$15,$1f,$00);
  AddAChar(58,$03,$00,$00,$00,$00,$00);
  AddAChar(59,$03,$00,$00,$00,$00,$00);
  AddAChar(60,$03,$00,$00,$00,$00,$00);
  AddAChar(61,$03,$00,$00,$00,$00,$00);
  AddAChar(62,$03,$00,$00,$00,$00,$00);
  AddAChar(63,$03,$00,$00,$00,$00,$00);
  AddAChar(64,$03,$00,$00,$00,$00,$00);
  AddAChar(65,$05,$07,$0a,$12,$0a,$07);
  AddAChar(66,$04,$1f,$15,$1f,$0a,$00);
  AddAChar(67,$04,$0e,$11,$11,$11,$00);
  AddAChar(68,$04,$1f,$11,$11,$0e,$00);
  AddAChar(69,$04,$1f,$15,$15,$11,$00);
  AddAChar(70,$04,$1f,$14,$14,$10,$00);
  AddAChar(71,$04,$0e,$11,$15,$17,$00);
  AddAChar(72,$04,$1f,$04,$04,$1f,$00);
  AddAChar(73,$03,$11,$1f,$11,$00,$00);
  AddAChar(74,$03,$12,$11,$1e,$00,$00);
  AddAChar(75,$04,$1f,$04,$0a,$11,$00);
  AddAChar(76,$03,$1f,$01,$01,$00,$00);
  AddAChar(77,$05,$1f,$08,$04,$08,$1f);
  AddAChar(78,$05,$1f,$08,$04,$02,$1f);
  AddAChar(79,$04,$0e,$11,$11,$0e,$00);
  AddAChar(80,$04,$1f,$14,$14,$08,$00);
  AddAChar(81,$04,$0e,$11,$12,$0d,$00);
  AddAChar(82,$04,$1f,$14,$16,$09,$00);
  AddAChar(83,$04,$09,$15,$15,$12,$00);
  AddAChar(84,$03,$10,$1f,$10,$00,$00);
  AddAChar(85,$04,$1e,$01,$01,$1e,$00);
  AddAChar(86,$05,$1c,$02,$01,$02,$1c);
  AddAChar(87,$05,$1f,$02,$04,$02,$1f);
  AddAChar(88,$05,$11,$0a,$04,$0a,$11);
  AddAChar(89,$05,$10,$08,$07,$08,$10);
  AddAChar(90,$05,$11,$13,$15,$19,$11);
  AddAChar(91,$03,$00,$00,$00,$00,$00);
  AddAChar(92,$03,$00,$00,$00,$00,$00);
  AddAChar(93,$03,$00,$00,$00,$00,$00);
  AddAChar(94,$03,$00,$00,$00,$00,$00);
  AddAChar(95,$03,$00,$00,$00,$00,$00);

  // Creating default scripts
  ScriptSet.Clear;
  AddScript(00,Script0000s,sizeof(Script0000s));
  AddScript(02,Script0001s,sizeof(Script0001s));
  AddScript(03,Script0001e,sizeof(Script0001e));
  AddScript(04,Script0010s,sizeof(Script0010s));
  AddScript(05,Script0010e,sizeof(Script0010e));
  AddScript(06,Script0011s,sizeof(Script0011s));
  AddScript(07,Script0011e,sizeof(Script0011e));
  AddScript(08,Script0100s,sizeof(Script0100s));
  AddScript(09,Script0100e,sizeof(Script0100e));
  AddScript(10,Script0101s,sizeof(Script0101s));
  AddScript(11,Script0101e,sizeof(Script0101e));
  AddScript(12,Script0110s,sizeof(Script0110s));
  AddScript(13,Script0110e,sizeof(Script0110e));
  AddScript(14,Script0111s,sizeof(Script0111s));
  AddScript(15,Script0111e,sizeof(Script0111e));
end;

// -----------------------------------------------------------------------------

procedure TSLV01.UpdateActiveScript;
var
  b,b1 : byte;
begin
  b := 0;
  if TPinInput(Inputs[0]).Value then b := b or $08;
  if TPinInput(Inputs[1]).Value then b := b or $02;
  if TPinInput(Inputs[2]).Value then b := b or $04;
  if (ActiveScript and b and $02) <> 0 then exit;
  if (ActiveScript and 1) = 0 then
  begin
    b1 := (ActiveScript xor b);
    if b1 <> 0 then
    begin
      if (b1 and b) <> 0 then
        ActiveScript := b
      else
        ActiveScript := ActiveScript or 1;
      ScriptSet[ActiveScript].Reset;
    end;
  end
  else
  begin
    ActiveScript := b;
    ScriptSet[ActiveScript].Reset;
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.Run(NoDelay : boolean; CS : TCommandSet;  Cmd : TGenericCommand) : TUndo;
var
  ii : integer;
  ascr : integer;
  RCmd : TGenericCommand;
  Row,Col : integer;
begin
  if NoDelay then
  begin
    ascr := ActiveScript;
    ActiveScript := ScriptSet.Add(CS);
    try
      ScriptSet[ActiveScript].ActiveCommand := 0;
      ii := 0;
      RCmd := nil;
      repeat
        if ii > ScriptSet[ActiveScript].ActiveCommand then break; // Loop
        if (ScriptSet[ActiveScript][ScriptSet[ActiveScript].ActiveCommand] is TSLV01Cmd_Stop) or
           (ScriptSet[ActiveScript][ScriptSet[ActiveScript].ActiveCommand] is TSLV01Cmd_Restart) then break;
        RCmd := ScriptSet[ActiveScript][ScriptSet[ActiveScript].ActiveCommand];
        for Row:=0 to PCB.RowCount-1 do
          for Col:=0 to PCB.ColCount-1 do
            result[Col,Row] := PCB[Col,Row].Active;
        if ScriptSet[ActiveScript].Run(PCB,CharTable,NoDelay) < 0 then break;
        PostMessage(frmMain.Handle,MSG_INVALIDATE,0,0);
        ii := ScriptSet[ActiveScript].ActiveCommand;
      until (RCmd = Cmd) and (RCmd <> ScriptSet[ActiveScript][ScriptSet[ActiveScript].ActiveCommand]);
    finally
      ScriptSet.Delete(ActiveScript);
      ActiveScript := ascr;
    end;
  end
  else
  begin
    result[0,0] := false;
    ii := ScriptSet[ActiveScript].Run(PCB,CharTable,NoDelay);
    if ii < 0 then
      UpdateActiveScript
    else
    begin
      result[0,0] := (ii > 1);
      if ((ActiveScript and 1) <> 0) and assigned(ScriptSet[ActiveScript].CurrentCommand) and (ScriptSet[ActiveScript].CurrentCommand is TSLV01Cmd_Stop) then
        UpdateActiveScript;
    end;
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.WriteTo(FileName : string; ID : integer) : integer;
var
  fs  : TFileStream;
  ss  : string;
  Buf : PByteArray;
  ii  : integer;
  bb  : byte;
  ch  : TCharacter;
  // ------------------------------------
  procedure DumpScript(ScrNo : integer);
  var
    Index : integer;
    Cmd : TGenericCommand;
    ii : integer;
  begin
    for ii:=0 to TopMem-1 do Buf[ii] := $ff;
    Index := 0;
    ScrNo := ScrNo shl 1;
    for ii:=0 to ScriptSet[ScrNo].Count-1 do
    begin
      Cmd := ScriptSet[ScrNo][ii];
      if (Cmd.Size+Index) > 1024 then exit;
      cmd.Save(Buf^,Index);
      if (cmd is TSLV01Cmd_Stop) or (cmd is TSLV01Cmd_Restart) then break;
    end;
    if ScrNo = 0 then exit;
    inc(ScrNo);
    Index := 512;
    for ii:=0 to ScriptSet[ScrNo].Count-1 do
    begin
      Cmd := ScriptSet[ScrNo][ii];
      if (Cmd.Size+Index) > 1024 then exit;
      cmd.Save(Buf^,Index);
      if (cmd is TSLV01Cmd_Stop) or (cmd is TSLV01Cmd_Restart) then break;
    end;
  end;
  // ------------------------------------
begin
  result := 0;
  try
    fs := TFileStream.Create(FileName,fmCreate);
    GetMem(Buf,TopMem);
    try
      ss := 'SLV01';
      fs.Write(ss[1],5);
      if ID = -1 then
      begin
        // Script Set
        ss := 'SET';
        fs.Write(ss[1],3);
        for ii:=0 to 7 do
        begin
          DumpScript(ii);
          fs.Write(Buf^,1024);
        end;
      end
      else
      if ID >= 0 then
      begin
        // One Script
        ss := format('S%2.2X',[ID]);
        fs.Write(ss[1],3);
        DumpScript(ID);
        fs.Write(Buf^,1024);
      end
      else
      begin
        // Char Table
        ss := 'CHT';
        fs.Write(ss[1],3);
        for ii:=0 to 999 do Buf[ii] := 0;
        for ii:=0 to 63 do
        begin
          ch := CharTable[32+ii];
          bb := length(ch.Data);
          Buf[ii*6] := bb;
          if bb > 0 then Buf[ii*6+1] := ch.Data[0];
          if bb > 1 then Buf[ii*6+2] := ch.Data[1];
          if bb > 2 then Buf[ii*6+3] := ch.Data[2];
          if bb > 3 then Buf[ii*6+4] := ch.Data[3];
          if bb > 4 then Buf[ii*6+5] := ch.Data[4];
        end;
        fs.Write(Buf^,384);
      end;
      result := 1;
    finally
      FreeMem(Buf);
      fs.Free;
    end;
  except
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.ReadFrom(FileName : string; ID : integer) : integer;
var
  fs  : TFileStream;
  Buf : PByteArray;
  ss  : string;
  bb  : byte;
  ii  : integer;
  Index : integer;
  sz,b1,b2,b3,b4,b5 : byte;
  // ----------------------------------------------
  function ParseScript(ScrNo : integer) : integer;
  var
    cs : TCommandSet;
    Index : integer;
    cmd : TGenericCommand;
  begin
    result := 0;
    ScrNo := ScrNo shl 1;
    cs := ScriptSet[ScrNo];
    cs.Clear;
    Index := 0;
    while Index < 1024 do
    begin
      cmd := TSLV01_LoadCommand(Buf^,Index);
      if Index > 1024 then exit;
      cs.Add(cmd);
      if (cmd is TSLV01Cmd_Stop) or (cmd is TSLV01Cmd_Restart) then
      begin
        if ((ScrNo and 1) = 0) and (ScrNo > 0) then
        begin
          inc(ScrNo);
          cs := ScriptSet[ScrNo];
          cs.Clear;
          Index := 512;
        end
        else
        begin
          break;
        end;
      end;
    end;
    result := 1;
  end;
  // ----------------------------------------------
  procedure AddAChar(Code,Len,b0,b1,b2,b3,b4 : byte);
  var
    ch : TCharacter;
    ii : integer;
  begin
    ch := CharTable[Code];
    if Len = 0 then exit;
    setlength(ch.Data,Len);
    for ii:=0 to Len-1 do
    begin
      case ii of
        0: ch.Data[ii] := b0;
        1: ch.Data[ii] := b1;
        2: ch.Data[ii] := b2;
        3: ch.Data[ii] := b3;
        4: ch.Data[ii] := b4;
      end;
    end;
  end;
  // ------------------------------------------------
begin
  result := 0;
  try
    fs := TFileStream.Create(FileName,fmOpenRead);
    GetMem(Buf,TopMem);
    try
      setlength(ss,8);
      if fs.Read(ss[1],8) < 8 then
      begin
        ShowMessage('Read Error');
        exit;
      end;
      if copy(ss,1,5) <> 'SLV01' then
      begin
        ShowMessage('This file is not for the currently selected device.');
        exit;
      end;
      ss := copy(ss,6,3);
      if ss = 'CHT' then
      begin
        if fs.Read(Buf^,384) <> 384 then
        begin
          ShowMessage('Read Error or File is too small');
          exit;
        end;
        CharTable.Clear;
        for Index:=0 to 63 do
        begin
          sz := Buf[Index*6];
          b1 := Buf[Index*6+1];
          b2 := Buf[Index*6+2];
          b3 := Buf[Index*6+3];
          b4 := Buf[Index*6+4];
          b5 := Buf[Index*6+5];
          AddAChar(32+Index,sz,b1,b2,b3,b4,b5);
        end;
      end
      else
      if ss = 'SET' then
      begin
        for ii:=0 to 7 do
        begin
          if fs.Read(Buf^,1024) <> 1024 then
          begin
            ShowMessage('Read Error or File is too small');
            exit;
          end;
          if ParseScript(ii) = 0 then
          begin
            ShowMessage('Unknown script command encountered');
            exit;
          end;
        end;
      end
      else
      begin
        bb := UnhexByte(copy(ss,2,2));
        if (ss[1] <> 'S') or (bb > 7) then
        begin
          ShowMessage('Can''t handle file of this type ('+ss+')');
          exit;
        end;
        if fs.Read(Buf^,1024) <> 1024 then
        begin
          ShowMessage('Read Error or File is too small');
          exit;
        end;
        if ParseScript(ID) = 0 then
        begin
          ShowMessage('Unknown script command encountered');
          exit;
        end;
      end;
      result := 1;
    finally
      FreeMem(Buf);
      fs.Free;
    end;
  except
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.GetScrLimit(ScrNo : byte) : integer;
begin
  if ScrNo = 0 then
    result := 1024
  else
    result := 512;
end;

// -----------------------------------------------------------------------------

function TSLV01.CmdFromText(ACmdText : string) : TGenericCommand;
begin
  result := TSLV01_CreateCommand(ACmdText);
end;

// -----------------------------------------------------------------------------

function TSLV01.WriteTo(FileName : string; Scr : TCommandSet) : integer;
var
  fs  : TFileStream;
  ss  : string;
  Buf : PByteArray;
  ii  : integer;
  Index : integer;
  Cmd : TGenericCommand;
begin
  result := 0;
  try
    fs := TFileStream.Create(FileName,fmCreate);
    GetMem(Buf,TopMem);
    try
      ss := 'SLV01';
      fs.Write(ss[1],5);
      // One Script
      ss := 'S00';
      fs.Write(ss[1],3);
      for ii:=0 to TopMem-1 do Buf[ii] := $ff;
      Index := 0;
      for ii:=0 to Scr.Count-1 do
      begin
        Cmd := Scr[ii];
        if (Cmd.Size+Index) > 1024 then exit;
        cmd.Save(Buf^,Index);
        if (cmd is TSLV01Cmd_Stop) or (cmd is TSLV01Cmd_Restart) then break;
      end;
      fs.Write(Buf^,1024);
      result := 1;
    finally
      FreeMem(Buf);
      fs.Free;
    end;
  except
  end;
end;

// -----------------------------------------------------------------------------

function TSLV01.ReadFrom(FileName : string; Scr : TCommandSet) : integer;
var
  fs  : TFileStream;
  Buf : PByteArray;
  ss  : string;
  Index : integer;
  cmd : TGenericCommand;
begin
  result := 0;
  try
    fs := TFileStream.Create(FileName,fmOpenRead);
    GetMem(Buf,TopMem);
    try
      setlength(ss,8);
      if fs.Read(ss[1],8) < 8 then
      begin
        ShowMessage('Read Error');
        exit;
      end;
      if copy(ss,1,5) <> 'SLV01' then
      begin
        ShowMessage('This file is not for the currently selected device.');
        exit;
      end;
      ss := copy(ss,6,3);
      if (ss[1] <> 'S') then
      begin
        ShowMessage('Can''t handle file of this type ('+ss+')');
        exit;
      end;
      if fs.Read(Buf^,1024) <> 1024 then
      begin
        ShowMessage('Read Error or File is too small');
        exit;
      end;
      Scr.Clear;
      Index := 0;
      while Index < 1024 do
      begin
        cmd := TSLV01_LoadCommand(Buf^,Index);
        if Index > 1024 then
        begin
          ShowMessage('Incorrect structure');
          exit;
        end;
        Scr.Add(cmd);
        if (cmd is TSLV01Cmd_Stop) or (cmd is TSLV01Cmd_Restart) then break;
      end;
      result := 1;
    finally
      FreeMem(Buf);
      fs.Free;
    end;
  except
  end;
end;

// -----------------------------------------------------------------------------

end.

