unit SerialIO;

interface

uses
  Windows, Classes, CBuf;

const
  SIO_ERR_NO_ERROR          = 0;
  SIO_ERR_CLOSED            = 1;
  SIO_ERR_CANT_OPEN_PORT    = 2;
  SIO_ERR_RX_BUFFER_OVERRUN = 3;
  SIO_ERR_IO_ERROR          = 4;

const
  SIO_CB_EVENT_BYTE_RECEIVED    = 1;
  SIO_CB_EVENT_CLEAR_TO_SEND    = 2;
  SIO_CB_EVENT_ERR_CODE_CHANGED = 3;

type
  TSerialIO = class;

  TSIOCallBack     = procedure (AEvent : integer; ASender : TSerialIO) of object;

  TSerialIO = class (TThread)
    constructor Create(Port : string; CBFunction : TSIOCallBack);
    destructor Destroy; override;
    procedure Execute; override;
    procedure Terminate; reintroduce;
  private
    FEvent   : integer;
    OvlRx    : OVERLAPPED;
    OvlTx    : OVERLAPPED;
    FErr     : integer;
    FWErr    : integer;
    FPort    : string;
    FCBFn    : TSIOCallBack;
    RTS      : boolean;
    Comm     : THandle;
    RxBuf    : TCBuf;
    TxBuf    : TCBuf;
    procedure Cleanup;
    procedure FnOnTerminate(Sender : TObject);
    function  GetDataSize : integer;
    procedure SIOCallBack(AEvent : integer);
  public
    iCRCNow : integer;
    property Error    : integer read FErr;
    property WinError : integer read FWErr;
    property DataSize : integer read GetDataSize;
    function Send(c : pchar; Len : integer) : integer;
    function Recv(c : pchar; Len : integer) : integer;
  end;

//function GetSIOStr(SIO : TSerialIO; var OutStr : string) : boolean;
function UnhexChar(cc : char) : byte;
function UnhexByte(ss : string) : byte;
function UnhexWord(ss : string) : word;

function HexByte(b : byte) : string;

function EnumerateComports(ComList : TStrings) : integer;

implementation

uses
  SysUtils, Forms;

// =============================================================================

function EnumerateComports(ComList : TStrings) : integer;
var
  ii : integer;
  ss : string;
  cc : COMMCONFIG;
  sz : dword;
begin
  result := 0;
  if not assigned(ComList) then exit;
  ComList.Clear;
  for ii:=1 to 255 do
  begin
    ss := 'COM'+IntToStr(ii);
    sz := sizeof(COMMCONFIG);
    if GetDefaultCommConfig(pchar(ss),cc,sz) then
    begin
      ComList.Add(ss+'='+cc.wcProviderData);
      inc(result);
    end;
  end;
end;

// =============================================================================

//function GetSIOStr(SIO : TSerialIO; var OutStr : string) : boolean;
//var
//  tout : integer;
//  ss   : string;
//  cc   : char;
//begin
//  result := false;
//  OutStr := '';
//  ss := '';
//  tout := 0;
//  repeat
//    if SIO.Recv(@cc,1) = 1 then
//    begin
//      if (byte(cc) >= $20) or ((cc in [#13,#10]) and (ss <> '')) then ss := ss+cc;
//      tout := 0;
//    end
//    else
//    begin
//      sleep(50);
//      inc(tout);
//    end;
//  until (tout >= 20) or (copy(ss,length(ss)-1,2) = #13#10);
//  if tout < 20 then
//  begin
//    for tout:=1 to length(ss) do
//      if byte(ss[tout]) >= $20 then
//        OutStr := OutStr+ss[tout];
//    result := true;
//  end;
//end;

// -----------------------------------------------------------------------------

function UnhexChar(cc : char) : byte;
begin
  if cc in ['0'..'9'] then
    result := (byte(cc)-$30)
  else
  if cc in ['a'..'f'] then
    result := (byte(cc)-$57)
  else
  if cc in ['A'..'F'] then
    result := (byte(cc)-$37)
  else
    result := 0;
end;

// -----------------------------------------------------------------------------

function UnhexByte(ss : string) : byte;
var
  ii : integer;
begin
  result := 0;
  for ii:=1 to length(ss) do result := result*16+UnhexChar(ss[ii]);
end;

// -----------------------------------------------------------------------------

function UnhexWord(ss : string) : word;
var
  ii : integer;
begin
  result := 0;
  for ii:=1 to length(ss) do result := result*16+UnhexChar(ss[ii]);
end;

// -----------------------------------------------------------------------------

const
  HexStr = '0123456789ABCDEF';

function HexByte(b : byte) : string;
begin
  result := HexStr[((b shr 4) and $0f)+1]+HexStr[(b and $0f)+1];
end;

// ========== TSerialIO ========================================================

constructor TSerialIO.Create(Port : string; CBFunction : TSIOCallBack);
begin
  inherited Create(true);
  iCRCNow := -1;
  FPort:=Port;
  FErr:=SIO_ERR_CLOSED;
  FWErr:=0;
  FCBFn:=CBFunction;
  RTS:=false;
  OnTerminate:=FnOnTerminate;
  FreeOnTerminate:=true;
  RxBuf:=TCBuf.Create(256);
  TxBuf:=TCBuf.Create(256);
  Resume;
end;

// -----------------------------------------------------------------------------

destructor TSerialIO.Destroy;
begin
  Cleanup;
  inherited Destroy;
end;

// -----------------------------------------------------------------------------

procedure TSerialIO.Cleanup;
begin
  FreeAndNil(RxBuf);
  FreeAndNil(TxBuf);
end;

// -----------------------------------------------------------------------------

procedure TSerialIO.FnOnTerminate(Sender : TObject);
begin
  FCBFn := nil;
  CloseHandle(Comm);
end;

// -----------------------------------------------------------------------------

procedure TSerialIO.Terminate;
begin
  FCBFn := nil;
  inherited Terminate;
end;

// -----------------------------------------------------------------------------

procedure TSerialIO.Execute;
var
  DCB  : TDCB;
  rxb  : byte;
  txb  : byte;
  rxdw : dword;
  txdw : dword;
  RIP  : boolean; // Read in progress
  TIP  : boolean; // Transmission in progress
  tout : COMMTIMEOUTS;
  bDoingSomething : boolean;
begin
  // Fill in the OVERLAPPED structures
  OvlRx.Internal     := 0;
  OvlRx.InternalHigh := 0;
  OvlRx.Offset       := 0;
  OvlRx.OffsetHigh   := 0;
  OvlRx.hEvent       := CreateEvent(nil,true,false,nil);
  OvlTx.Internal     := 0;
  OvlTx.InternalHigh := 0;
  OvlTx.Offset       := 0;
  OvlTx.OffsetHigh   := 0;
  OvlTx.hEvent       := CreateEvent(nil,true,false,nil);
  // Creating the communication file
  Comm:=CreateFile(PChar('\\.\'+FPort),GENERIC_WRITE or GENERIC_READ,0,nil,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL or FILE_FLAG_OVERLAPPED,0);
  if Comm = INVALID_HANDLE_VALUE then
  begin
    FErr:=SIO_ERR_CANT_OPEN_PORT;
    FWErr:=GetLastError;
    exit;
  end;
  GetCommState(Comm,DCB);
  with DCB do
  begin
    Flags     := 3;
    BaudRate  := CBR_9600;
    ByteSize  := 8;
    Parity    := NOPARITY;
    StopBits  := ONESTOPBIT;
    DCBlength := SizeOf(DCB);
  end;
  tout.ReadIntervalTimeout         := 1000;
  tout.ReadTotalTimeoutMultiplier  := 300;
  tout.ReadTotalTimeoutConstant    := 1000;
  tout.WriteTotalTimeoutConstant   := 0;
  tout.WriteTotalTimeoutMultiplier := 0;
  if not SetCommTimeouts(Comm,tout) then
  begin
    FErr:=SIO_ERR_CANT_OPEN_PORT;
    FWErr:=GetLastError;
    CloseHandle(Comm);
    exit;
  end;
  if not SetCommState(Comm,DCB) then
  begin
    FErr:=SIO_ERR_CANT_OPEN_PORT;
    FWErr:=GetLastError;
    CloseHandle(Comm);
    exit;
  end;
  if not SetCommMask(Comm,EV_BREAK or EV_ERR or EV_RXCHAR or EV_TXEMPTY) then
  begin
    FErr:=SIO_ERR_CANT_OPEN_PORT;
    FWErr:=GetLastError;
    CloseHandle(Comm);
    exit;
  end;
  FErr:=SIO_ERR_NO_ERROR;
  RIP:=false;
  TIP:=false;
  // Working...
  bDoingSomething := True;
  while not Terminated do
  begin
    if not bDoingSomething then
      sleep(1);
    bDoingSomething := False;
    // Attempting to start the Read Operation
    if not RIP then
    begin
      if ReadFile(Comm,rxb,1,rxdw,@OvlRx) then
      begin
        if (rxdw = 1) then
        begin
          bDoingSomething := True;
          if (RxBuf.Put(@rxb,1) <> 1) then
          begin
            FErr:=SIO_ERR_RX_BUFFER_OVERRUN;
            SIOCallBack(SIO_CB_EVENT_ERR_CODE_CHANGED);
          end
          else
            SIOCallBack(SIO_CB_EVENT_BYTE_RECEIVED);
        end;
      end
      else
      begin
        case GetLastError of
          ERROR_IO_PENDING: RIP:=true;
          ERROR_HANDLE_EOF:; // Just ignore it
        else
          FErr:=SIO_ERR_IO_ERROR;
          SIOCallBack(SIO_CB_EVENT_ERR_CODE_CHANGED);
        end;
      end;
    end;
    // Attempting to start the Write Operation
    if (TxBuf.DataSize > 0) and (not TIP) and (TxBuf.Get(@txb,1) = 1) then
    begin
      if not WriteFile(Comm,txb,1,txdw,@OvlTx) then
      begin
        if GetLastError = ERROR_IO_PENDING then
          TIP:=true
        else
        begin
          FErr:=SIO_ERR_IO_ERROR;
          SIOCallBack(SIO_CB_EVENT_ERR_CODE_CHANGED);
        end;
      end
      else
      begin
        bDoingSomething := True;
//        if (TxBuf.DataSize = 0) then
//          SIOCallBack(SIO_CB_EVENT_CLEAR_TO_SEND);
      end;
    end;
    if RIP then
    begin
      if GetOverlappedResult(Comm,OvlRx,rxdw,false) then
      begin
        RIP:=false;
        if (rxdw = 1) then
        begin
          if (RxBuf.Put(@rxb,1) <> 1) then
          begin
            FErr:=SIO_ERR_RX_BUFFER_OVERRUN;
            SIOCallBack(SIO_CB_EVENT_ERR_CODE_CHANGED);
          end
          else
            SIOCallBack(SIO_CB_EVENT_BYTE_RECEIVED);
          bDoingSomething := True;
        end;
      end
      else
      begin
        if GetLastError <> ERROR_IO_INCOMPLETE then
        begin
          RIP:=false;
          FErr:=SIO_ERR_IO_ERROR;
          SIOCallBack(SIO_CB_EVENT_ERR_CODE_CHANGED);
        end;
      end;
    end;
    if TIP then
    begin
      if GetOverlappedResult(Comm,OvlTx,txdw,false) then
      begin
        TIP:=false;
        if txdw = 0 then
        begin
          if not WriteFile(Comm,txb,1,txdw,@OvlTx) then
          begin
            if GetLastError = ERROR_IO_PENDING then
              TIP:=true
            else
            begin
              FErr:=SIO_ERR_IO_ERROR;
              SIOCallBack(SIO_CB_EVENT_ERR_CODE_CHANGED);
            end;
          end;
        end
        else
        begin
          bDoingSomething := True;
//          if (TxBuf.DataSize = 0) then
//            SIOCallBack(SIO_CB_EVENT_CLEAR_TO_SEND);
        end;
      end
      else
      begin
        if GetLastError <> ERROR_IO_INCOMPLETE then
        begin
          TIP:=false;
          FErr:=SIO_ERR_IO_ERROR;
          SIOCallBack(SIO_CB_EVENT_ERR_CODE_CHANGED);
        end
        else
          bDoingSOmething := True;
      end;
    end;
  end;
end;

// -----------------------------------------------------------------------------

function TSerialIO.Send(c : pchar; Len : integer) : integer;
begin
  result:=TxBuf.Put(c,Len);
end;

// -----------------------------------------------------------------------------

function TSerialIO.Recv(c : pchar; Len : integer) : integer;
begin
  result:=RxBuf.Get(c,Len);
end;

// -----------------------------------------------------------------------------

function TSerialIO.GetDataSize : integer;
begin
  result:=RxBuf.DataSize;
end;

// -----------------------------------------------------------------------------

procedure TSerialIO.SIOCallBack(AEvent: integer);
begin
  if assigned(FCBFn) then FCBFn(AEvent,self)
end;

// =============================================================================

end.

